# This package contains general functions used by Stanford Perl programs, to
# standardize usage.

package Stanford::Infrared::General;

use 5.010;
use autodie;
use strict;
use warnings;

use Config::Simple;
use DB_File;
use Fcntl qw(LOCK_EX LOCK_NB);
use JSON;
use Net::Remctl;
use Readonly;

# Functions we export on request.  Separated out into related groups.
use base 'Exporter';
our @EXPORT_OK = qw(
    get_db_setup format_quarter get_next_quarters

    valid_directories dir_exists has_write_access get_pts_admins

    notify

    workgroup_members pts_members

    map_servers

    lock_queue unlock_queue

    validate_remctl
    );

# Servers to contact for remctl commands.
Readonly my $NOTIFY => 'devnull.stanford.edu';
Readonly my $LSDB   => 'lsdb.stanford.edu';

##########################################################################
# Exception routines
##########################################################################

# Usage      : validate_remctl($result, $status_die, $message)
# Purpose    : Pulls many checks for remctl errors to one place, checking the
#              error and standard error, and dying if there was a failure.
# Parameters : $result     - Result of a Net::Remctl call
#              $status_die - Die on a non-zero status
#              $message    - Descriptive text to print on error
# Throws     : String exception on any critical errors
sub validate_remctl {
    my ($result, $status_die, $message) = @_;

    if ($result->error) {
        die "remctl failure for '$message': " . $result->error . "\n";
    }

    if ($status_die && $result->status != 0) {
        if ($result->stderr) {
            die "remctl error for '$message': " . $result->stderr . "\n";
        } else {
            die "remctl error for '$message'\n";
        }
    }
}

##########################################################################
# General parsing routines
##########################################################################

# Usage      : get_db_setup($fname)
# Purpose    : Read a standard config file with database information, as we
#              use on tools and other servers.  Do some parsing of various
#              alternative fields used in the past, then find the information
#              required to create a connection.
# Returns    : $dsn      - DSN string used for the connection.  This includes
#                          the host, port, and database, so they are usually
#                          not actually needed, but returned just in case.
#              $username - Username to authenticate as
#              $password - Password to authenticate as
#              $host     - Host to connect to
#              $port     - Server port to connect to
#              $database - Database name to use
# Parameters : $fname - Name of the configuration file to read from
# Throws     : String exception on file not found
#              String exception on any data missing from file
sub get_db_setup {
    my ($fname) = @_;

    die "no database config file: $fname\n" unless -e $fname;

    # Get the values from config file.
    my $cnf = Config::Simple->new ($fname);
    my $database = $cnf->param ('database') || $cnf->param ('db_name');
    my $host     = $cnf->param ('host') || $cnf->param ('db_host');
    my $port     = $cnf->param ('port') || $cnf->param ('db_port');
    my $username = $cnf->param ('username') || $cnf->param ('db_username');
    my $password = $cnf->param ('password') || $cnf->param ('db_password');

    # Die if any are missing.
    die "no database given in database config file\n" unless $database;
    die "no host given in database config file\n"     unless $host;
    die "no port given in database config file\n"     unless $port;
    die "no username given in database config file\n" unless $username;
    die "no password given in database config file\n" unless $password;

    # Create the dsn and return everything.
    my $dsn = sprintf ("DBI:mysql:database=%s;host=%s;port=%d", $database,
                       $host, $port);

    return ($dsn, $username, $password, $host, $port, $database);
}

# Usage      : format_quarter($quarter)
# Purpose    : Translates a Peoplesoft quarter into something more suitable
#              for the general public to read, for such things as menus.
# Returns    : Quarter in readable format, like "Fall Quarter 2010-2011"
# Parameters : $quarter - Peoplesoft-formatted quarter
sub format_quarter {
    my ($quarter) = @_;

    my %quartermap = ( '2' => 'Fall Quarter',
                       '4' => 'Winter Quarter',
                       '6' => 'Spring Quarter',
                       '8' => 'Summer Quarter' );
    $quarter =~ /^\d(\d\d)(\d)$/;
    return sprintf ("%s %d-%d", $quartermap{$2}, $1 + 1999, $1 + 2000);
}

# Usage      : get_next_quarters($max_quarters)
# Purpose    : Find a number of quarters that we can then use to present a
#              user with a listing of quarters to select from, for services
#              tied to a specific quarter or quarter expiration.
# Returns    : Hash of quarters, with keys the Peoplesoft quarter numbers and
#              values a version meant for human readability.
# Parameters : $max_quarters - The number of quarters to return
sub get_next_quarters {
    my ($max_quarters) = @_;

    # A map of months to what quarter we consider to be our first starting
    # quarter for a date in that month.  This was generated by looking at the
    # academic calendar for normal dates and assuming that someone will want
    # to create a course volume for the current quarter up til the month it
    # ends.  Start with January as 0.
    my @QUARTERMONTHS = (4, 4, 4, 6, 6, 8, 8, 8, 2, 2, 2, 2);

    # Get the current term based on this month.  If we're on autumn term,
    # then we're actually at next school year, so increment the year.
    my (undef, $min, $hour, $mday, $mon, $shortyear) = localtime;
    my $term = $QUARTERMONTHS[$mon];
    $shortyear =~ s#^\d*(\d{2})$#$1#;
    $shortyear++ if $term == 2;

    # Find the next several quarters.  When we go past summer quarter, start
    # over with the next school year.
    my %quarters;
    for (my $i = 0; $i < $max_quarters; $i++) {
        my $psquarter = sprintf ("1%02d%d", $shortyear, $term);
        my $string= '('.$psquarter.')  '. format_quarter ($psquarter);
        $quarters{$psquarter} = $string;

        $term += 2;
        if ($term > 8) {
            $term = 2;
            $shortyear++;
        }
    }

    return %quarters;
}

##########################################################################
# AFS-related routines
##########################################################################

# These all require afsdir remctl access on lsdb, which generally only
# service/tools and individual users have.

# Usage      : valid_directories($sunetid)
# Purpose    : Contact lsdb to get a list of all valid group, department, and
#              class directories that a user is in the -admins PTS group for.
# Returns    : 1 if access
#              0 if none
# Parameters : $sunetid - Sunetid to check for access
# Throws     : String exception on remctl failure
sub valid_directories {
    my ($sunetid) = @_;
    my $result = remctl ($LSDB, 0, '', 'afsdir', 'directories', $sunetid);
    validate_remctl ($result, 1, "$LSDB afsdir directories $sunetid");

    return () unless $result->stdout;
    my @dirs = split (/[\r\n]+/, $result->stdout);
    return sort @dirs;
}

# Usage      : dir_exists($dir)
# Purpose    : Contact lsdb to use its privileges to see if an AFS dir exists
# Returns    : 1 if exists
#              0 if not
# Parameters : $dir     - AFS directory to check against
# Throws     : String exception on remctl failure
sub dir_exists {
    my ($dir) = @_;
    my $result = remctl ($LSDB, 0, '', 'afsdir', 'chkexist', $dir);
    validate_remctl ($result, 0, "$LSDB afsdir chkexist $dir");

    return 1 if $result->status == 0;
    return 0;
}

# Usage      : has_write_access($sunetid, $dir)
# Purpose    : Contact lsdb to see if a user has write access to a given
#              directory.
# Returns    : 1 if access
#              0 if none
# Parameters : $sunetid - Sunetid to check for write access
#              $dir     - AFS directory to check against
# Throws     : String exception on remctl failure
sub has_write_access {
    my ($sunetid, $dir) = @_;
    my $result = remctl ($LSDB, 0, '', 'afsdir', 'chkwrite', $dir, $sunetid);
    validate_remctl ($result, 0, "$LSDB afsdir chkwrite $dir $sunetid");

    return 1 if $result->status == 0;
    return 0;
}

# Usage      : get_pts_admins($org)
# Purpose    : Find the members of an AFS organization's owning PTS group
# Returns    : Sorted list of sunetids
# Parameters : $org - AFS organization, in the format of the group, class, or
#              department home directory without leading /afs/ir/
# Throws     : String exception on invalid organization
#              String exception on remctl failure
#              String exception on remctl non-zero return value
sub get_pts_admins {
    my ($org) = @_;

    # Map the organization to an admins group.
    my $pts;
    if ($org =~ m#^group/(.+)$#) {
        $pts = $1.'-admins';
    } elsif ($org =~ m#^dept/(.+)$#) {
        $pts = 'dept_'.$1.'-admins';
    } elsif ($org =~ m#^class/archive/[^/]+/[^/]+/([^/]+)\.(\d+)$#) {
        $pts = $2.'-'.$1.'-admins';
    } else {
        die "invalid organization: $org\n";
    }

    # Contact the remctl server and die on any errors.
    my $result = remctl ($LSDB, 0, '', 'afsdir', 'membership', $pts);
    validate_remctl ($result, 1, "$LSDB afsdir membership $pts");

    # Return the users.  No stdout means no valid users rather than error.
    return () if !defined $result->stdout;
    my @users = split (/[\r\n]+/, $result->stdout);
    return sort @users;
}

##########################################################################
# Mail-related routines
##########################################################################

# Usage      : notify($dryrun, %msgvars)
# Purpose    : Use the notifications service to send or print an email.
#              The service and what it normally expects are documented in our
#              wiki.
# Returns    : 1 on success
#              0 on failure
# Parameters : $dryrun  - Set true to print the mail, not send
#              %msgvars - Hash of keys and values to pass to the service
# Throws     : String exception on remctl failure
sub notify {
    my ($dryrun, %msgvars) = @_;
    my $command = 'mail-json';
    $command = 'print-json' if $dryrun;

    my $json = JSON->new;
    my $json_text = $json->encode (\%msgvars);
    my $result = remctl ($NOTIFY, 0, '', 'notifications', $command,
                         $json_text);
    validate_remctl ($result, 0, "$NOTIFY notifications $command DATA");
    warn $result->stderr if $result->stderr;

    print $result->stdout if $dryrun && $result->stdout;

    return 1 if $result->status == 0;
    return 0;
}

##############################################################################
# Membership/owner handling
##############################################################################

# Usage      : workgroup_members($group, $dir)
# Purpose    : Check and return the membership of a workgroup.  We assume
#              that we already have gotten any needed Kerberos permissions,
#              as those can vary.
# Returns    : Sorted list of the sunetids in the given group
# Parameters : $group - Name of the PTS group
#              $dir   - Stanford::Directory connection already made
sub workgroup_members {
    my ($workgroup, $dir) = @_;
    return () if !$workgroup;

    my @entries = $dir->ldap_query ("suprivilegegroup=$workgroup");
    if (!@entries && $dir->error) {
        warn "Error in workgroup_members: " . $dir->error_type . " "
            . $dir->error . "\n";
    }

    my @users;
    foreach my $entry (@entries) {
        foreach my $attr (keys (%{$entry})) {
            if ($attr eq 'uid') {
                foreach (@{$entry->{$attr}}) {
                    push(@users, $_);
                }
            }
        }
    }

    return sort @users;
}

# Usage      : pts_members($group)
# Purpose    : Check and return the membership of a PTS group.  We assume
#              that we already have gotten any needed AFS permissions, as
#              those can vary.
# Returns    : Sorted list of the sunetids in the given group
# Parameters : $group - Name of the PTS group
# TODO:        This is currently used by tools and the workgroup integration
#              engine.  Perhaps we instead want a version that contacts lsdb
#              like get_pts_admins above.
sub pts_members {
    my ($group) = @_;
    return () if !$group;

    my $pts = AFS::PTS->new;
    my @members = $pts->members ($group);
    return sort @members;
}

##############################################################################
# Group and user mappings
##############################################################################

# Usage      : map_servers ($map_fname, $skip_types)
# Purpose    : Open a mapping of servers to sysadmins and groups, then use
#              those mappings to find connections between users and groups
#              and users and servers.
# Returns    : hashref mapping groups to hosts
#              hashref mapping primary sysadmins to hosts
#              hashref mapping secondary/tertiary/etc sysadmins to hosts
#              hashref mapping sunetids to groups
# Parameters : $map_fname      - file name to read for the server mappings
#              $skip_types_ref - hashref of any types to skip (usually
#                                windows servers)
# Throws     : String exception on not being able to find file
sub map_servers {
    my ($map_fname, $skip_types_ref) = @_;

    my %servers;
    die "could not find $map_fname\n" unless -f $map_fname;
    tie %servers, 'DB_File', $map_fname;

    my (%groups, %primary, %secondary, %sunetid_to_group);
    foreach my $host (keys %servers) {
        my ($group, $type, $primary, @secondaries)
            = split (/,/, $servers{$host});
        next unless defined $type;
        next if $type && exists $skip_types_ref->{$type};

        $host .= '.stanford.edu' unless $host =~ /\./;
        $group   ||= 'NONE';
        $primary ||= 'NONE';

        push (@{$groups{lc $group}}, $host);
        push (@{$primary{lc $primary}}, $host);
        foreach my $admin (@secondaries) {
            push (@{$secondary{$admin}}, $host);
            $sunetid_to_group{$admin} = lc $group;
        }
        $sunetid_to_group{$primary} = lc $group;
    }
    untie %servers;

    return (\%groups, \%primary, \%secondary, \%sunetid_to_group);
}

# Usage      : server_access_group ($server, $group, %groups)
# Purpose    : Using a map of groups to servers returned by map_servers,
#              check to see if a given server is in the list of servers
#              supported by a group.
# Returns    : 1 if the server is in the group's list
#              0 if it is not, or if there is no matching group
# Parameters : $server - Server to check ownership for.
#              $group  - Group to check ownership for.
#              %groups - Hash of groups to lists of servers.
sub server_access_group {
    my ($server, $group, %groups) = @_;
    return 0 if !exists $groups{$group};

    my @servers = @{$groups{$group}};
    for my $check_server (@servers) {
        return 1 if $check_server eq $server;
    }
    return 0;
}

##############################################################################
# Queue handling
##############################################################################

# Usage      : lock_queue($lockfile, $max_lockchecks, $rest)
# Purpose    : Create a lockfile for a program, to avoid multiple runs of a
#              program at once.  Attempt multiple times, in case another
#              short-lived instance is running.
# Returns    : 1 on success
#              undef on failure
# Parameters : $lockfile       - Filename of the lockfile to use
#              $max_lockchecks - Maximum number of times to try locking
#              $rest           - Seconds to rest between checks
# Throws     : I/O exception on file open failure
#              String exception on failure to lock after $max_lockchecks
sub lock_queue {
    my ($lockfile, $max_lockchecks, $rest) = @_;
    my $count = 0;
    open (LOCKFILE, '>', $lockfile);
    while ($count < $max_lockchecks) {
        if (flock (LOCKFILE, LOCK_EX | LOCK_NB)) {
            return 1;
        } else {
            $count++;
            sleep $rest unless ($count == $max_lockchecks);
        }
    }

    die "unable to lock lockfile after $count attempts\n";
}

# Usage      : unlock_queue($lockfile)
# Purpose    : Remove and delete an existing lockfile.
# Returns    : Return value of the unlink
# Parameters : $lockfile       - Filename of the lockfile to use
# Throws     : I/O exception on file remove failure
sub unlock_queue {
    my ($lockfile) = @_;
    close LOCKFILE;
    unlink $lockfile;
}

1;

__END__

##############################################################################
# Documentation
##############################################################################

=head1 NAME

Stanford::Infrared::General - Miscellaneous functions for infrastructure

=head1 SYNOPSIS

    use Stanford::Infrared::General qw(get_db_setup);

    my ($dsn, $username, $password) = get_db_setup('/etc/databases/proxydb');

=head1 DESCRIPTION

This module includes general commands useful for various infrastructure
machines.  These include standard data parsers, queue handling, queries
on the db file of systems to owners, and sending mail via the
notifications service.

=head1 COMMANDS

=over 4

=item B<get_db_setup> ($fname)

Reads in a standard formatted database setup file, with lines including
the servername, database, username, and password.  Creates and returns
a DSN line, the username, and password.  This is used to centralize
processing for a number of scripts that read in database information
from a puppetized file.

=item B<format_quarter> ($quarter)

Takes a Peoplesoft formatted quarter (such as 1106) and returns
something more understandable by a person (such as "Spring Quarter
2011-2012").

=item B<get_next_quarters> ($max_quarters)

Generates and returns a hash of the next several quarters, given by the
$max_quarters variable.  The hash keys are the Peoplesoft quarter, and
the values are the expanded name via B<format_quarter>.  This is used
to generate a list of upcoming quarters for services that need to offer
users a choice of quarters.

=item B<valid_directories> ($sunetid)

Returns a list of valid AFS groups, departments, and classes that a
user has access to via the user's membership in a PTS admins group.
Used to help generate a list of AFS organizations a user has the right
to do things for.

=item B<dir_exists> ($directory)

Returns 1 if the given directory exists, 0 if not.  Calls lsdb via
remctl to use its permissions to get a canonical answer.

=item B<has_write_access> ($sunetid, $directory)

Returns 1 if the user has write access to the given directory, 0 if
not.  Calls lsdb via remctl to use its permissions to get a canonical
answer.

=item B<get_pts_admins> ($organization)

Returns a list of sunetids in the PTS admins group for an organization.
Calls lsdb via remctl to use its permissions to get a canonical answer.

=item B<notify> ($dryrun, %message_vars)

Calls the Notifications service and asks it to send out a mail.
%message_vars should be loaded with the service and message name, along
with any variables that should be sent.  When $dryrun is set, only
print the message to the screen.

=item B<workgroup_members> ($workgroup, $dir_handle)

Find and return a list of the members of a given workgroup, via a
pre-created Stanford::Directory handle.  Assumes you already have
permissions set up to do so.

=item B<pts_members> ($ptsgroup)

Find and return the members of a given PTS group.  Assumes you already
have permissions set up to do so.

=item B<map_servers> ($fname, $skip_types_ref)

Read in a Berkeley DB file that contains a mapping of server names to
groups, types, and the sysadmins that run them.  Optionally skip
certain types (usually Windows servers) that you don't want listed.

=item B<lock_queue> ($lockfile, $max_lockchecks, $rest_between_attempts)

Create a lock file, for services that need checking to make sure that a
long-lived process via cron isn't having another process run over it.

=item B<unlock_queue> ($lockfile)

Removes a lock created previously by lock_queue.

=item B<validate_remctl> ($result, $status_die, $message)

Checks a Net::Remctl result to see if we should die for any critical
errors running the command.  If so, print out the given error message
along with surrounding context.  If status_die is set, also die on any
non-zero status from the remctl command.

=back

=head1 AUTHORS

Jon Robertson <jonrober@stanford.edu>

=cut
