# This package contains general functions used by Stanford Perl programs
# doing various functions against our standard database remctl commands.

package Stanford::Infrared::MySQL;

use 5.010;
use autodie;
use strict;
use warnings;

use Readonly;
use Net::Remctl;
use Stanford::Infrared::General qw(validate_remctl);

use base 'Exporter';
our @EXPORT_OK = qw(store_database_login generate_db_username
    generate_afs_db_name org_under_db_quota create_database
    create_afs_database create_afs_account db_passwd_from_file);

# Server to run privileged AFS commands against.
Readonly my $LSDB => 'lsdb.stanford.edu';

##########################################################################
# Internal functions
##########################################################################

# Usage      : _generate_db_prefix($org_type)
# Purpose    : Finds the database prefix assigned for a type of database
# Returns    : Prefix to assign to every database of the given type
# Parameters : $org_type - Organizational type (AFS directory or 'drupal')
sub _generate_db_prefix {
    my ($org_type) = @_;

    my $prefix = '';
    if    ($org_type eq 'class')  { $prefix = 'c'  }
    elsif ($org_type eq 'dept')   { $prefix = 'd'  }
    elsif ($org_type eq 'group')  { $prefix = 'g'  }
    elsif ($org_type eq 'drupal') { $prefix = 'dr' }

    return $prefix;
}

##########################################################################
# MySQL service routines
##########################################################################

# Usage      : store_database_login($afsdir, $sunetid_owner, $db_name,
#                                   $db_admin_name, $password, $server)
# Purpose    : Add database login information to the file for an AFS
#              organization.  If the directory doesn't exist or is
#              inaccessible, contact lsdb to create it.  We assume the caller
#              already has service/mysql permissions.
# Returns    : 1 if success
#              0 if failure
# Parameters : $afsdir        - Home directory for the owning org
#              $sunetid_owner - User who requested the database
#              $db_name       - Database name
#              $db_admin_name - Username for the database
#              $password      - Password for the database
#              $server        - MySQL server the db was created on
# Throws     : String exception on remctl failure
sub store_database_login {
    my ($afsdir, $sunetid_owner, $db_name, $db_admin_name, $password,
        $server) = @_;

    # Create db_private directory if it does not already exist.
    if (! -e $afsdir . '/db_private/') {
        my $result = remctl ($LSDB, 0, '', 'mysql', 'create', $afsdir);
        validate_remctl ($result, 1, "$LSDB mysql create $afsdir");
    }

    # Add database information to the info file.
    my $db_info    = $afsdir . '/db_private/db_info';
    my $created_on = localtime;
    my $status;
    if (open (DB_INFO, '>>', $db_info)) {
        print DB_INFO "\n";
        print DB_INFO "MySQL username: $db_admin_name\n";
        print DB_INFO "MySQL initial password: $password\n";
        print DB_INFO "Database server: $server\n";
        print DB_INFO "Database name: $db_name\n";
        print DB_INFO "Requested by: $sunetid_owner\n";
        print DB_INFO "Created-on: $created_on\n";
        close DB_INFO;
        return 1;
    } else {
        warn "$0: failed to open $db_info: $!\n";
        return 0;
    }
}

# Usage      : generate_db_username($db_server, $db_name, $read_only)
# Purpose    : Generate a unique database username from the database name
# Returns    : Unique username
# Parameters : $db_server - Database server to check for names on
#              $db_name   - Database name to use to generate username
#              $read_only - Flag for a read-only account (ends in _ro)
# Throws     : String exception on remctl failure
sub generate_db_username {
    my ($db_server, $db_name, $read_only) = @_;
    $read_only = 0 unless defined $read_only;
    my $username = $db_name;

    # Clean up the possible username to remove special characters and
    # standardize.
    $username =~ s#[^\w]##;
    $username =~ s#_##g;
    $username =~ s#\.##g;
    $username = lc ($username);

    # Shorten the username, as mysql has 16-char limit.  Read-only needs
    # extra space for '_ro'.
    if ($read_only) {
        $username =~ s#^(.{11}).+$#$1#;
    } else {
        $username =~ s#^(.{14}).+$#$1#;
    }

    # Make sure the name is unique.
    my $extension = 0;
    my $tmpname = $username;
    $tmpname .= '_ro' if $read_only;
    my $result = remctl ($db_server, 0, '', 'db', 'user', $tmpname);
    validate_remctl ($result, 0, "$db_server db user $tmpname");

    while ($result->status == 0) {
        $extension++;
        $tmpname = $username . $extension;
        $tmpname .= '_ro' if $read_only;
        $result = remctl ($db_server, 0, '', 'db', 'user', $tmpname);
        validate_remctl ($result, 0, "$db_server db user $tmpname");
    }

    return $tmpname;
}

# Usage      : generate_afs_db_name($db_server, $type, $main_name)
# Purpose    : Generate a unique database name for a given server
# Returns    : Database name
# Parameters : $db_server - Server to check against for unique name
#              $org_type  - Type of AFS directory (dept|class|group)
#              $org_name  - Name of AFS directory
#              $main_name - User-requested or service-generated name, last
#                           part of actual name
# Throws     : String exception on remctl failure
sub generate_afs_db_name {
    my ($db_server, $org_type, $org_name, $main_name) = @_;

    # Calculate the default database name.
    my $prefix = _generate_db_prefix ($org_type);
    my $db_name = lc join ('_', $prefix, $org_name, $main_name);
    $db_name =~ s#[/\.-]#_#g;

    # Check to see if the calculated database name exists.  If so, add
    # numbers to the end until we find one that doesn't.  The account names
    # should be unique and so this shouldn't be a problem.
    my $extension = '';
    my $result = remctl ($db_server, 0, '', 'db', 'check', $db_name);
    validate_remctl ($result, 0, "$db_server db check $db_name");

    while ($result->status == 0) {
        $extension = 0 if $extension eq '';
        $extension++;
        $result = remctl ($db_server, 0, '', 'db', 'check',
                          $db_name.$extension);
        validate_remctl ($result, 0,
                         "$db_server db check ${db_name}${extension}");
    }
    $db_name .= $extension;

    return $db_name;
}

# Usage      : org_under_db_quota($db_server, $org_type, $org_name)
# Purpose    : Contact a database server to see if an AFS organization can
#              create more databases.
# Returns    : 1 if allowed
#              0 if not
# Parameters : $db_server - Database server to contact
#              $org_type  - Organizational type (dept|class|group)
#              $org_name  - Organizational name
# Throws     : String exception on remctl failure
sub org_under_db_quota {
    my ($db_server, $org_type, $org_name) = @_;

    # Generate the prefix for all databases for this org.
    my $db_prefix = _generate_db_prefix ($org_type);
    $db_prefix .= '_' . $org_name;
    $db_prefix =~ s{-}{_}g;

    # Check to see if the organization is allowed more databases.
    my $result = remctl ($db_server, 0, '', 'db', 'allowed', $db_prefix);
    validate_remctl ($result, 0, "$db_server db allowed $db_prefix");

    if ($result->status != 0) {
        return 0;
    }

    return 1;
}

# Usage      : create_database($db_server, $db_name, $username, $password)
# Purpose    : Create a database on a given mysql server via remctl
# Returns    : password for the database
# Parameters : $db_server - Database server to create on
#              $db_name   - Name of the database to create
#              $username  - Username for the database
#              $password  - (optional) Password for the db.  Assigned by the
#                           mysql server if not given.
# Throws     : String exception on any remctl failures
sub create_database {
    my ($db_server, $db_name, $username, $password) = @_;

    # Create the database and user.  If we were given a password, then pass
    # it along to set that exact password.
    my ($result);
    if ($password) {
        $result = remctl ($db_server, 0, '', 'db', 'create', '--passwd',
                          $password, $db_name, $username);
    } else {
        $result = remctl ($db_server, 0, '', 'db', 'create', $db_name,
                          $username);
    }
    validate_remctl ($result, 1, "$db_server db create $db_name $username");

    if ($result->status == 0) {
        $password = $result->stdout;
        chomp $password;
        return $password;
    }
}

# Usage      : create_afs_database ($db_server, $org_type, $org_name,
#                                   $db_barename, $owner, $afsdir)
# Purpose    : Perform all tasks for creating a database for an AFS
#              organization.  This includes creating the database from basic
#              info, writing the file to AFS containing login info, and
#              adding an entry to the sumysql tracking table.
# Returns    : Array of database name, database username, and password.
# Parameters : $db_server   - Database server to create against
#              $org_type    - AFS organization type (dept|class|group)
#              $org_name    - Name of the AFS subdirectory
#              $db_barename - Name of the actual requested database itself
#              $owner       - Requesting sunetid
# Throws     : String exception on any failure to create database
#              String exception on remctl failure
sub create_afs_database {
    my ($db_server, $org_type, $org_name, $db_barename, $owner) = @_;

    # AFS directory generated from org.
    my $afsdir = sprintf ("/afs/ir/%s/%s/", $org_type, $org_name);

    # Get default values and create the database.
    my $db_name   = generate_afs_db_name ($db_server, $org_type,
                                          $org_name, $db_barename);
    die "could not generate database name\n" unless $db_name;
    my $username = generate_db_username ($db_server, $db_name, 0);
    die "could not generate username\n" unless $username;
    my $password = create_database ($db_server, $db_name, $username, '');
    die "could not create database\n" unless $password;

    # Now store the db information in a file.
    my $status = store_database_login ($afsdir, $owner, $db_name, $username,
                                       $password, $db_server);
    warn "could not store database login\n" unless $status;

    # Now add the record to the sumysql tracking table.
    my $result = remctl ($db_server, 0, '', 'db', 'sumysql-insert',
                         $org_type, $org_name, $db_name, $username, $owner);
    validate_remctl ($result, 1, "$db_server db sumysql-insert");

    return ($db_name, $username, $password);
}

# Usage      : create_afs_account ($db_server, $db_name, $db_user, $db_pass,
#                                  $owner, $org_type, $org_name, $read_only)
# Purpose    : Creates a new user account with access to a database.
# Returns    : Array of username and password on success
# Parameters : $db_server - Database server to create against
#              $db_name   - Database name to give access to
#              $db_user   - Database user name to create
#              $db_pass   - (optional) If not '', password to use for account
#              $owner     - sunetid of requesting user
#              $org_type  - Type of organization owning db (group/dept/class)
#              $org_name  - Name of the organization owning db (afs name)
#              $read_only - If set, create account as read-only
# Throws     : String exception on any failure to create database
#              String exception on remctl failure
sub create_afs_account {
    my ($db_server, $db_name, $db_user, $db_pass, $owner, $org_type,
        $org_name, $read_only) = @_;

    # AFS directory generated from org.
    my $afsdir = sprintf ("/afs/ir/%s/%s/", $org_type, $org_name);

    # Generate a username for this database.
    $db_user = generate_db_username ($db_server, $db_name, $read_only);

    # Create the account.  If we were given a password, then pass it along to
    # set that exact password.
    my ($result);
    my $flag = '';
    $flag = '--readonly' if $read_only;
    if ($db_pass) {
        $result = remctl ($db_server, 0, '', 'db', 'create-account',
                          '--passwd', $flag, $db_pass, $db_name, $db_user);
    } else {
        $result = remctl ($db_server, 0, '', 'db', 'create-account', $flag,
                          $db_name, $db_user);
    }
    validate_remctl ($result, 1,
                     "$db_server db create-account $db_name $db_user");

    # Get the returned password if the command was successful, or die.
    if ($result->status == 0) {
        ($db_user, $db_pass) = split /\s/, $result->stdout;
        chomp $db_pass;
    } else {
        die "could not create account $db_name\n";
    }

    # Now store the db information in a file.
    my $status = store_database_login ($afsdir, $owner, $db_name, $db_user,
                                       $db_pass, $db_server);
    warn "could not store database login\n" unless $status;

    # Add the record to the sumysql tracking table.
    $result = remctl ($db_server, 0, '', 'db', 'sumysql-insert',
                      $org_type, $org_name, $db_name, $db_user, $owner);
    validate_remctl ($result, 1, "$db_server db sumysql-insert");

    return ($db_user, $db_pass);
}

# Usage      : db_passwd_from_file
# Purpose    : Open a file containing only a password and return the
#              contents.  Used with our normal process of writing new db
#              passwords to disk.
# Returns    : Password
# Parameters : $fname - Name of file to read from
# Throws     : I/O exceptions on open or close failure
sub db_passwd_from_file {
    my ($fname) = @_;
    open (PASSWD, '<', $fname);
    my $passwd = <PASSWD>;
    close PASSWD;
    chomp $passwd;
    return $passwd;
}

1;

__END__

##############################################################################
# Documentation
##############################################################################

=head1 NAME

Stanford::Infrared::MySQL - MySQL functions for infrastructure machines

=head1 SYNOPSIS

    use Stanford::Infrared::MySQL qw(create_afs_database);
    my ($db_name, $username, $password)
      = create_afs_database ('mysql05', 'group', 'web-test', 'new', 'rra');

=head1 DESCRIPTION

This module includes various commands for creating and maintaining
MySQL databases against the Stanford infrastructure.  It assumes that
the servers it's run against will be using the stanford-server-mysql
Debian package with its remctl interfaces, and in large part exists as
a client wrapper for calls to those functions.

=head1 COMMANDS

=over 4

=item B<store_database_login> ($afsdir, $sunetid_owner, $db_name,
$db_admin_name, $password, $server)

This stores login information in an AFS directory, under the db_private
directory there.  The AFS directory should be the home directory for a
group, department, or class.  Login information is added to the db_info
file in the directory, preserving any earlier entries.

=item B<generate_db_username> ($db_server, $db_name, $read_only)

Generates a unique database username on a server, using given database
name as a base.  The database name is truncated if too long, and then
_ro is appended if it is for a read-only user account.  After that, the
server is contacted to make sure the name is unique, and if not then
the database name is munged until it is unique.

=item B<generate_afs_db_name> ($db_server, $org_type, $org_name,
$main_name)

For a database that's meant to be strongly tied to an AFS group (such
as those on the mysql-user service), generate a database name.  We take
the owning AFS organization and their requested name, then use remctl
to find a unique unused version of the name.

=item B<org_under_db_quota> ($db_server, $org_type, $org_name)

Checks to see if the given organization is under their quota and can
still create databases.  Returns 1 if true, 0 if not.

=item B<create_database> ($db_server, $db_name, $username, $password)

Create a new database on the given server, with the given information.
If the password is not given, one will be generated by the server and
returned.

=item B<create_afs_database> ($db_server, $org_type, $org_name,
$db_barename, $owner)

Create a database strongly tied to an AFS account.  This generates and
returns the database name, username, and password.

=item B<create_afs_account> ($db_server, $db_name, $db_user, $db_pass,
$owner, $org_type, $org_name, $read_only)

Creates an AFS account, separate of the one created when a database is
created.  This is usually used for creating a read-only secondary
account for a database.

=item B<db_passwd_from_file> ($fname)

Reads and returns a single line from a file.  In this case it's meant
to be a file containing only a database password.

=back

=head1 AUTHORS

Jon Robertson <jonrober@stanford.edu>

=cut
