# This package contains functions used to wrapper normal Perl functions, for
# internal usability or flexibility.

package Stanford::Infrared::Wrappers;

use 5.010;
use autodie;
use strict;
use warnings;

use IPC::Open3;
use Carp;

# Functions we export on request.  Separated out into related groups.
use base 'Exporter';
our @EXPORT_OK = qw(unquote_heredoc print_fh say_fh print_stdout say_stdout
    system_output);

#############################################################################
# Documentation-related routines
#############################################################################

# This strips formatting off of a string.  The purpose for it is to set up
# heredocs with extra formatting that makes them more readable in the midst
# of other code, then run through this function in order to clean them up
# into a normal heredoc.  It assumes that all the lines of the heredoc begin
# with a colon and seven spaces (used to set them off from the start of
# line.
#
# Example usage:
#
#    print {$mail} unquote_heredoc(<<"EOM");
#:       From: $maintainer
#:       To: $from
#:       Subject: comp.lang.perl.moderated registration confirmation
#:
#:       You have been registered as a poster for comp.lang.perl.moderated.
#:       All future posts from the addresses:
#:
#EOM
#
# $string - String to de-colonify and remove spaces from.
#
# Returns: String without the extra formatting
sub unquote_heredoc {
    my ($string) = @_;
    $string =~ s{ ^ : (?: [ ]{0,7} | \t ) }{}xmsg;
    return $string;
}

#############################################################################
# Output-related routines
#############################################################################

# This is a wrapper around print, to properly do checking of whether the
# print failed and die if so.  It's used since autodie doesn't check print,
# and we should just in case of rare problems.
#
# $fh    - Previously set up filehandle to print to
# @print - Array of items to give to print
#
# Returns: undef
#  Throws: dies with text if cannot write to the filehandle
sub print_fh {
    my ($fh, @print) = @_;
    print {$fh} @print or croak "print to a filehandle failed: $!";
    return;
}

# This is a wrapper around say, to properly do checking of whether the say
# failed and die if so.  It's used since autodie doesn't check say, and we
# should just in case of rare problems.
#
# $fh    - Previously set up filehandle to print to
# @print - Array of items to give to print
#
# Returns: undef
#  Throws: dies with text if cannot write to the filehandle
sub say_fh {
    my ($fh, @say) = @_;
    say {$fh} @say or croak "say to a filehandle failed: $!";
    return;
}

# This is a wrapper around print, to properly do checking of whether the
# print failed and die if so.  It's used since autodie doesn't check print,
# and we should just in case of rare problems.
#
# This version assumes stdout rather than asking for a filehandle.
#
# @print - Array of items to give to print
#
# Returns: undef
#  Throws: dies with text if cannot write to the filehandle
sub print_stdout {
    my (@print) = @_;
    print {*STDOUT} @print or croak "print to STDOUT failed: $!";
    return;
}

# This is a wrapper around say, to properly do checking of whether the say
# failed and die if so.  It's used since autodie doesn't check say, and we
# should just in case of rare problems.
#
# This version assumes stdout rather than asking for a filehandle.
#
# @print - Array of items to give to print
#
# Returns: undef
#  Throws: dies with text if cannot write to the filehandle
sub say_stdout {
    my (@say) = @_;
    say {*STDOUT} @say or croak "say to STDOUT failed: $!";
    return;
}

#############################################################################
# Systems-related routines
#############################################################################

# This is a wrapper around the IPC::Open3 open3 command, that retrieves
# both the exit status and output of a program, returning both.  It's used
# for situations where we would normally run with backticks, so that we can
# also examine the exit status.
#
# @prog - Program and arguments to run, given as a list.
#
# Returns: exit status
#          string of output from program
#  Throws: dies with text on unable to run open3 or error from child proc
sub system_output {
    my (@prog) = @_;

    # Run the program.
    my ($in, $out, $err);
    my $pid = open3($in, $out, $err, @prog)
        or croak "$0: error on open3 $prog[0]: $!\n";
    waitpid($pid, 0);
    my $exit_status = $? >> 8;

    # Get a select to read in the output responses.
    my $sel = new IO::Select;
    $sel->add($out);

    # Now grab all the output together into a string.
    my $all_out = '';
    while (my @ready = $sel->can_read) {
        foreach my $fh (@ready) {
            my $data;
            read $fh, $data, 8196;
            if (!defined $data) {
                croak "error from child: $!\n";
            }
            if (length $data == 0) {
                $sel->remove($fh);
                next;
            } else {
                $all_out .= $data;
            }
        }
    }

    return ($exit_status, $all_out);
}

1;

__END__

##############################################################################
# Documentation
##############################################################################

=head1 NAME

Stanford::Infrared::Wrappers - Various flexibility wrappers around functions

=head1 SYNOPSIS

    use Stanford::Infrared::MySQL qw(print_stdout system_output);

    print_stdout ("This is a test\n");

    my ($status, $output) = system_output('/usr/bin/foobar', '-l');

=head1 DESCRIPTION

This module includes various commands that wrapper Perl fucntions for
stability and flexibility.  They include greater error checking and
needed functionality in addition to the standards.

=head1 COMMANDS

=over 4

=item B<unquote_heredoc> ($string)

Removes special quoting around a heredoc, as used by Russ to help them
be more readable when in the middle of other code.

=item B<print_fh> ($fh, @print_lines)

Print to a filehandle, with error checking and croak if the print
fails.

=item B<say_fh> ($fh, @print_lines)

say to a filehandle, with error checking and croak if the say fails.

=item B<print_stdout> (@print_lines)

Print to STDOUT, with error checking and croak if the print fails.

=item B<say_stdout> (@print_lines)

say to STDOUT, with error checking and croak if the say fails.

=item B<system_output> (@command)

Run a command and return both the return status of the run and any
STDOUT.

=back

=head1 AUTHORS

Jon Robertson <jonrober@stanford.edu>

=cut
